//############################################################################
//
// LaserBoy !!!
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
// Copyright 2003, 2004, 2005, 2006, 2007, 2008, 2009 James Lehman.
// This source is distributed under the terms of the GNU General Public License.
//
// LaserBoy_TUI.cpp is part of LaserBoy.
//
// LaserBoy is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// LaserBoy is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LaserBoy. If not, see <http://www.gnu.org/licenses/>.
//
//############################################################################
#include "LaserBoy_GUI_base.hpp"

//############################################################################
void LaserBoy_TUI::overlay_palettes()
{
    int    i,
           j,
           y_line       = 8 * font_size_factor,
           y_off        =   LASERBOY_PALETTE_MAX * font_size_factor
                          + LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           x_off        = bmp.yres + LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);
    //------------------------------------------------------------------------
    if(palette_index != current_frame().palette_index)
        palette_index = current_frame().palette_index;
    //------------------------------------------------------------------------
    if(palette_index != LASERBOY_TRUE_COLOR) // 8 bit frame palette
    {
        for(i = 0; i < const_palette_picker(palette_index).number_of_colors(); i++)
            for(j = 0; j < font_size_factor; j++)
                bmp_put_line_fast(&bmp,
                                  x_off,
                                  i * font_size_factor + j,
                                  x_off + 50,
                                  i * font_size_factor + j,
                                  bmp.make_rgb(const_palette_picker(palette_index)[i].r,
                                               const_palette_picker(palette_index)[i].g,
                                               const_palette_picker(palette_index)[i].b
                                              )
                                 );
    }
    //------------------------------------------------------------------------
    if(palette_index == LASERBOY_TRUE_COLOR || show_target_palette)
        for(i = 0; i < const_palette_picker(target_palette_index).number_of_colors(); i++) // target palette
            for(j = 0; j < font_size_factor; j++)
                bmp_put_line_fast(&bmp,
                                  x_off + 70,
                                  i * font_size_factor + j,
                                  x_off + 120,
                                  i * font_size_factor + j,
                                  bmp.make_rgb(const_palette_picker(target_palette_index)[i].r,
                                               const_palette_picker(target_palette_index)[i].g,
                                               const_palette_picker(target_palette_index)[i].b
                                              )
                                 );
    //------------------------------------------------------------------------
    if(show_color_cursor) // selected_color_index example line
    {
        if(palette_index != LASERBOY_TRUE_COLOR)
            bmp_put_line_fast(&bmp,
                              x_off + 56,
                              0,
                              x_off + 56,
                              LASERBOY_PALETTE_MAX * font_size_factor,
                              bmp.make_rgb(const_palette_picker(palette_index)[selected_color_index].r,
                                           const_palette_picker(palette_index)[selected_color_index].g,
                                           const_palette_picker(palette_index)[selected_color_index].b
                                          )
                             );
    //------------------------------------------------------------------------
        if(palette_index == LASERBOY_TRUE_COLOR || show_target_palette)
            bmp_put_line_fast(&bmp,
                              x_off + 63,
                              0,
                              x_off + 63,
                              LASERBOY_PALETTE_MAX * font_size_factor,
                              bmp.make_rgb(const_palette_picker(target_palette_index)[selected_color_index].r,
                                           const_palette_picker(target_palette_index)[selected_color_index].g,
                                           const_palette_picker(target_palette_index)[selected_color_index].b
                                          )
                             );
    //------------------------------------------------------------------------
        for(j = 0; j < font_size_factor; j++) // selected color marker
            bmp_put_line_fast(&bmp,
                              x_off + 40,
                              selected_color_index * font_size_factor + j,
                              x_off + 80,
                              selected_color_index * font_size_factor + j,
                              bmp.white
                             );
    }
    //------------------------------------------------------------------------
    if(palette_index == LASERBOY_TRUE_COLOR)
        bmp_printf(   &bmp // current frame palette name
                    , x_off
                    , y_off
                    , bmp_bg_color
                    , bmp.white
                    , 0
                    , font_size_factor
                    , "-24-bit-"
                  );
    else
        bmp_printf(   &bmp
                    , x_off
                    , y_off
                    , bmp.white
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , const_palette_picker(palette_index).name.c_str()
                  );
    //------------------------------------------------------------------------
    if(palette_index == LASERBOY_TRUE_COLOR || show_target_palette)
        bmp_printf(   &bmp // target palette name
                    , x_off + 130
                    , 0
                    , bmp.white
                    , bmp_bg_color
                    , 1
                    , font_size_factor
                    , const_palette_picker(target_palette_index).name.c_str()
                  );
    //------------------------------------------------------------------------
    if(show_color_cursor)
    {
        if(palette_index == LASERBOY_TRUE_COLOR || show_target_palette)
        {
            bmp_printf(   &bmp // target palette name
                        , x_off + 130
                        , y_line * 10
                        , bmp.make_rgb(0,0,255)
                        , bmp_bg_color
                        , 1
                        , font_size_factor
                        , "            %3d"
                        , const_palette_picker(target_palette_index)[selected_color_index].b
                      );
            bmp_printf(   &bmp // target palette name
                        , x_off + 130
                        , y_line * 10
                        , bmp.make_rgb(0,255,0)
                        , bmp_bg_color
                        , 1
                        , font_size_factor
                        , "        %3d"
                        , const_palette_picker(target_palette_index)[selected_color_index].g
                      );
            bmp_printf(   &bmp // target palette name
                        , x_off + 130
                        , y_line * 10
                        , bmp.make_rgb(255,0,0)
                        , bmp_bg_color
                        , 1
                        , font_size_factor
                        , "    %3d"
                        , const_palette_picker(target_palette_index)[selected_color_index].r
                      );
            bmp_printf(   &bmp // target palette name
                        , x_off + 130
                        , y_line * 10
                        , bmp.white
                        , bmp_bg_color
                        , 1
                        , font_size_factor
                        , "%3d"
                        , selected_color_index
                      );
        }
        y_off += y_line;
        bmp_printf(   &bmp // selected color blue value
                    , x_off
                    , y_off
                    , bmp.make_rgb(0,0,255)
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "            %3d"
                    , const_palette_picker(palette_index)[selected_color_index].b
                  );
        bmp_printf(   &bmp // selected color green value
                    , x_off
                    , y_off
                    , bmp.make_rgb(0,255,0)
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "        %3d"
                    , const_palette_picker(palette_index)[selected_color_index].g
                  );
        bmp_printf(   &bmp // selected color red value
                    , x_off
                    , y_off
                    , bmp.make_rgb(255,0,0)
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "    %3d"
                    , const_palette_picker(palette_index)[selected_color_index].r
                  );
        bmp_printf(   &bmp // selected color index
                    , x_off
                    , y_off
                    , bmp.white
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "%3d"
                    , selected_color_index
                  );
    }
}

//############################################################################
void LaserBoy_TUI::render_space()
{
    LaserBoy_color  bit_masked;

    u_int           select_start_color,
                    select_end_color  ,
                    bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    int             x_off        = bmp.yres + LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
                    y_off        =   LASERBOY_PALETTE_MAX * font_size_factor
                                   + (LASERBOY_STATS_DISPLAY_OFFSET + 24) * font_size_factor,
                    y_line       = 8 * font_size_factor;

    double          scale        = bmp.yres / 65536.0;
    //------------------------------------------------------------------------
    render_frame(frame_index);
    //------------------------------------------------------------------------
    if(show_palette || show_target_palette)
        overlay_palettes();
    //------------------------------------------------------------------------
    if(show_stats)
    {
        if(current_frame().format == LASERBOY_2D_FRAME)
        {
            if(current_frame().is_selected)
                bmp_printf(   &bmp
                            , x_off
                            , y_off
                            , bmp.make_rgb(255,255,0)
                            , bmp.make_rgb(128,0,0)
                            , 0
                            , font_size_factor
                            , "2D frame %6d"
                            , frame_index
                          );
            else
                bmp_printf(   &bmp
                            , x_off
                            , y_off
                            , bmp.make_rgb(255,255,0)
                            , bmp_bg_color
                            , 0
                            , font_size_factor
                            , "2D frame %6d"
                            , frame_index
                          );
        }
        else
        {
            //----------------------------------------------------------------
            if(current_frame().is_selected)
                bmp_printf(   &bmp
                            , x_off
                            , y_off
                            , bmp.make_rgb(255,255,0)
                            , bmp.make_rgb(128,0,0)
                            , 0
                            , font_size_factor
                            , "3D frame %6d"
                            , frame_index
                          );
            else
                bmp_printf(   &bmp
                            , x_off
                            , y_off
                            , bmp.make_rgb(255,255,0)
                            , bmp_bg_color
                            , 0
                            , font_size_factor
                            , "3D frame %6d"
                            , frame_index
                          );
        }
        //--------------------------------------------------------------------
        y_off += y_line;
        if(current_frame().is_unique)
            bmp_printf(   &bmp
                        , x_off
                        , y_off
                        , bmp.make_rgb(0,255,255)
                        , bmp_bg_color
                        , 0
                        , font_size_factor
                        , "vertices %6d"
                        , current_frame().quantity
                      );
        else
            bmp_printf(   &bmp
                        , x_off
                        , y_off
                        , bmp.make_rgb(0,255,255)
                        , bmp.make_rgb(0,128,0)
                        , 0
                        , font_size_factor
                        , "vertices %6d"
                        , current_frame().quantity
                      );
        //--------------------------------------------------------------------
        y_off += y_line;
        bmp_printf(   &bmp
                    , x_off
                    , y_off
                    , bmp.make_rgb(255,0,255)
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "segments %6d"
                    , current_frame().number_of_segments()
                  );
    } // end if(show_stats)
    if(current_frame().size() && show_cursors)
    {
        bit_masked = current_frame().at(current_frame().egg).bit_masked_color(signal_bit_mask);
        select_start_color =   current_frame().at(current_frame().egg).is_blank()
                             ? bmp.gray
                             :   current_frame().at(current_frame().egg).is_black(black_level)
                               ? bmp.white
                               : bmp.make_rgb(bit_masked.r, bit_masked.g, bit_masked.b);
        //----------------------------------------------------------------
        bit_masked = current_frame().at(current_frame().spider).bit_masked_color(signal_bit_mask);
        select_end_color   =   current_frame().at(current_frame().spider).is_blank()
                             ? bmp.gray
                             :   current_frame().at(current_frame().spider).is_black(black_level)
                               ? bmp.white
                               : bmp.make_rgb(bit_masked.r, bit_masked.g, bit_masked.b);
        //----------------------------------------------------------------
        // icons
        y_off += (2 * y_line); // color vertex segment for start cursor
        bmp_printf(   &bmp
                    , x_off
                    , y_off
                    , bmp.make_rgb(255,0,255)
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "           %4d"
                    , current_frame().segment_index_of_vertex(current_frame().egg)
                  );
        bmp_printf(   &bmp
                    , x_off
                    , y_off
                    , bmp.make_rgb(0,255,255)
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "      %4d"
                    , current_frame().egg
                  );
        bmp_printf(   &bmp
                    , x_off
                    , y_off
                    , bmp.white
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "  %3d"
                    , current_frame().at(current_frame().egg).c
                  );
        //----------------------------------------------------------------
        y_off += y_line; // RGB values for start cursor
        bmp_printf(   &bmp
                    , x_off
                    , y_off
                    , bmp.make_rgb(0,0,255)
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "            %3d"
                    , current_frame().at(current_frame().egg).b
                  );
        bmp_printf(   &bmp
                    , x_off
                    , y_off
                    , bmp.make_rgb(0,255,0)
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "        %3d"
                    , current_frame().at(current_frame().egg).g
                  );
        bmp_printf(   &bmp
                    , x_off
                    , y_off
                    , bmp.make_rgb(255,0,0)
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "    %3d"
                    , current_frame().at(current_frame().egg).r
                  );
        //----------------------------------------------------------------
        y_off += y_line; // color vertex segment for end cursor
        bmp_printf(   &bmp
                    , x_off
                    , y_off
                    , bmp.make_rgb(255,0,255)
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "           %4d"
                    , current_frame().segment_index_of_vertex(current_frame().spider)
                  );
        bmp_printf(   &bmp
                    , x_off
                    , y_off
                    , bmp.make_rgb(0,255,255)
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "      %4d"
                    , current_frame().spider
                  );
        bmp_printf(   &bmp
                    , x_off
                    , y_off
                    , bmp.white
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "  %3d"
                    , current_frame().at(current_frame().spider).c
                  );
        //----------------------------------------------------------------
        y_off += y_line; // RGB values for end cursor
        bmp_printf(   &bmp
                    , x_off
                    , y_off
                    , bmp.make_rgb(0,0,255)
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "            %3d"
                    , current_frame().at(current_frame().spider).b
                  );
        bmp_printf(   &bmp
                    , x_off
                    , y_off
                    , bmp.make_rgb(0,255,0)
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "        %3d"
                    , current_frame().at(current_frame().spider).g
                  );
        bmp_printf(   &bmp
                    , x_off
                    , y_off
                    , bmp.make_rgb(255,0,0)
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "    %3d"
                    , current_frame().at(current_frame().spider).r
                  );
        //-----------------------------------------------------------------
        y_off -= (3 * y_line);
        bmp_put_rectangle(&bmp, x_off
                              , y_off
                              , x_off + (int)(2000 * scale)
                              , y_off + (int)(2000 * scale)
                              , select_start_color
                              , 0
                         ); // start cursor icon
        //----------------------------------------------------------------
        y_off += (2 * y_line);
        bmp_put_line_fast     (&bmp, x_off + (int)(1000 * scale)
                                   , y_off
                                   , x_off + (int)(1000 * scale)
                                   , y_off + (int)(2000 * scale)
                                   , select_end_color
                              );
        bmp_put_line_fast     (&bmp, x_off + (int)(2000 * scale)
                                   , y_off + (int)(1000 * scale)
                                   , x_off
                                   , y_off + (int)(1000 * scale)
                                   , select_end_color
                              );
        bmp_put_line_fast     (&bmp, x_off
                                   , y_off
                                   , x_off + (int)(2000 * scale)
                                   , y_off + (int)(2000 * scale)
                                   , select_end_color
                              );
        bmp_put_line_fast     (&bmp, x_off + (int)(2000 * scale)
                                   , y_off
                                   , x_off
                                   , y_off + (int)(2000 * scale)
                                   , select_end_color
                              ); // end cursor icon
    //------------------------------------------------------------------------
    } // end if(show_cursors)
    //------------------------------------------------------------------------
    if(    show_TUI_clue
        && TUI_clue.size()
      )
    {
        bmp_printf(   &bmp
                    , bmp.yres + LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor
                    , bmp.yres - 8 * font_size_factor
                    , bmp.make_rgb(255,255,0)
                    , bmp.make_rgb(63,0,63)
                    , 0
                    , font_size_factor
                    , "%s                                                     "
                    , TUI_clue.c_str()
                  );
    }
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_frame_stats()
{
    int   vertices     = current_frame().quantity,
          lit          = current_frame().number_of_color_vertices(),
          blank        = current_frame().number_of_dark_vertices(),
          y_line       = 8 * font_size_factor,
          x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
          y_off        = bmp.yres - 14 * y_line; // 19 lines from the bottom
    u_int bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    LaserBoy_vertex    masked;
    //------------------------------------------------------------------------
    bmp_printf(&bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor,
               "frame stats"
              );
    y_off += y_line;
    bmp_printf(&bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor,
               "   lit vertices %d ~ %0.2f%%",
               lit,
               (lit/(double)vertices) * 100.0
              );
    y_off += y_line;
    bmp_printf(&bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor,
               "  dark vertices %d ~ %0.2f%%",
               blank,
               (blank/(double)vertices) * 100.0
              );
    y_off += y_line;
    bmp_printf(&bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor,
               " max lit vector %0.2f points",
               current_frame().max_color_distance()
              );
    y_off += y_line;
    bmp_printf(&bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor,
               " max drk vector %0.2f points",
               current_frame().max_dark_distance()
              );
    y_off += y_line;
    bmp_printf(&bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor,
               "   total length %0.2f points",
               current_frame().total_distance()
              );
    y_off += y_line;
    bmp_printf(&bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor,
               "     avg vector %0.2f (total length / vertices)",
               (current_frame().total_distance() / vertices)
              );
    y_off += y_line;
    bmp_printf(&bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor,
               " sharpest angle %0.2f%c",
               (current_frame().max_angle() * one_radian),
               0xf8
              );
    y_off += y_line;
    bmp_printf(&bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor,
               "    total angle %0.2f%c",
               (current_frame().total_angle() * one_radian),
               0xf8
              );
    if(show_cursors)
    {
        masked = current_frame().at(current_frame().egg).bit_masked(signal_bit_mask);
        if(current_frame().palette_index != LASERBOY_TRUE_COLOR)
        {
            masked.r = palette_picker(current_frame().palette_index).at(current_frame().at(current_frame().egg).c).r;
            masked.g = palette_picker(current_frame().palette_index).at(current_frame().at(current_frame().egg).c).g;
            masked.b = palette_picker(current_frame().palette_index).at(current_frame().at(current_frame().egg).c).b;
        }
        y_off += y_line;
        bmp_printf(&bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor,
                   "%c %s i=%6d X=%7d Y=%7d Z=%7d r=%3d g=%3d b=%3d",
                   0x01,
                   ((current_frame().at(current_frame().egg).is_lit()) ? ("lit  ") : ("blank")),
                   current_frame().egg,
                   masked.x,
                   masked.y,
                   masked.z,
                   masked.r,
                   masked.g,
                   masked.b
                  );
        masked = current_frame().at(current_frame().spider).bit_masked(signal_bit_mask);
        if(current_frame().palette_index != LASERBOY_TRUE_COLOR)
        {
            masked.r = palette_picker(current_frame().palette_index).at(current_frame().at(current_frame().spider).c).r;
            masked.g = palette_picker(current_frame().palette_index).at(current_frame().at(current_frame().spider).c).g;
            masked.b = palette_picker(current_frame().palette_index).at(current_frame().at(current_frame().spider).c).b;
        }
        y_off += y_line;
        bmp_printf(&bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor,
                   "%c %s i=%6d X=%7d Y=%7d Z=%7d r=%3d g=%3d b=%3d",
                   0x0f,
                   ((current_frame().at(current_frame().spider).is_lit()) ? ("lit  ") : ("blank")),
                   current_frame().spider,
                   masked.x,
                   masked.y,
                   masked.z,
                   masked.r,
                   masked.g,
                   masked.b
                  );
        y_off += y_line;
        bmp_printf(&bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor,
                   "%c-%c cntr X=%0.2f Y=%0.2f Z=%0.2f diff=%0.2f",
                   0x01,
                   0x0f,
                   current_frame().selected_segment().find_center().x,
                   current_frame().selected_segment().find_center().y,
                   current_frame().selected_segment().find_center().z,
                     current_frame().at(current_frame().egg) | current_frame().at(current_frame().spider)
                  );
    }
    if(show_fulcrum)
    {
        y_off += y_line;
        bmp_printf(&bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor,
                   "    fulcrum X=%0.2f Y=%0.2f Z=%0.2f",
                   fulcrum.x,
                   fulcrum.y,
                   fulcrum.z
                  );
    }
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_frame_set_stats()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = bmp.yres + LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = bmp.yres - 7 * y_line;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);
    count_frames();
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.white
                , bmp_bg_color
                , 0
                , font_size_factor
                , "frame set stats"
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , next_hue()
                , bmp_bg_color
                , 0
                , font_size_factor
                , "frames    %d"
                , number_of_frames()
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , next_hue()
                , bmp_bg_color
                , 0
                , font_size_factor
                , "2D frames %d"
                , num_2D_frames
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , next_hue()
                , bmp_bg_color
                , 0
                , font_size_factor
                , "3D frames %d"
                , num_3D_frames
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , next_hue()
                , bmp_bg_color
                , 0
                , font_size_factor
                , "verticies %d"
                , total_vertices()
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_palette_set_stats()
{
    int    i,
           y_line       = 8 * font_size_factor,
           x_off        = bmp.yres + LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp
                , x_off
                , 0
                , bmp.white
                , bmp_bg_color
                , 0
                , font_size_factor
                , "                  "
              );
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.white
                , bmp_bg_color
                , 0
                , font_size_factor
                , "palette set stats "
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , next_hue()
                , bmp_bg_color
                , 0
                , font_size_factor
                , "total palettes %d "
                , number_of_palettes()
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , next_hue()
                , bmp_bg_color
                , 0
                , font_size_factor
                , "added palettes %d"
                , number_of_palettes() - LASERBOY_BUILT_IN_PALETTES
              );
    for(i = 0; i < number_of_palettes(); i++)
    {
        y_off += y_line;
        bmp_printf(   &bmp
                    , x_off
                    , y_off
                    , next_hue()
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "%s          "
                    , LaserBoy_palette_set::at(i).name.c_str()
                  );
    }
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_program_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "Welcome to " LASERBOY_VERSION " !!!"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    i o input | output file"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      t signal bit resolution"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      u user interface visuals"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      v view keys & apply"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      x system switch settings"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [Tab] system value  settings"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    ` ~ show"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    - _ orbit"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    = + show & orbit"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    %c %c frame"
                , 0x1b
                , 0x1a
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    %c %c target palette"
                , 0x18
                , 0x19
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    9 0 first | last frame"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    ( ) set | unset frames-( | )-frames select"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "[Space] select | unselect current frame"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "[Bkspc] omit current frame"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      p palette   transforms"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      h frame     transforms"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      j frame set transforms"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      b wave file utilities"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      k draw, color, blank  vertices"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      l move, scale, rotate segments"
              );
    y_off += y_line;
    if(bg_file_name != ".")
    {
        y_off += y_line;
        bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "      ; show | hide background bitmap"
                  );
    }
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    ? / show | hide frame | frame set stats"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[Enter] show | hide this menu"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "  [Esc] exit LaserBoy"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_in_file_type_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "file input type"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 ild"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 dxf"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    3 wav"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    4 txt"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    5 bmp"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    6 wtf"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[Esc] cancel action"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_in_ild_action_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "ild file input action"
              );

    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 replace frame set"
              );

    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 prepend to frame set"
              );

    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    3 insert after current frame"
              );

    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    4 append to frame set"
              );

    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    5 superimpose frame(s) into frame set"
              );

    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[Esc] cancel action"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_in_dxf_action_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "DXF file input action"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "frame"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 replace frame set with dxf file"
              );

    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 append  frame set with dxf file"
              );

    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "frame set"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    3 replace frame set with dxf directory"
              );

    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    4 append  frame set with dxf directory"
              );

    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[Esc] cancel action"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_in_wav_action_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "wave file input action"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "LaserBoy formatted wave"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 replace current frame set"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 append to current frame set"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    3 open as unformatted wave"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "unformatted wave"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    4 replace current frame set"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    5 append to current frame set"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    6 reframe based on blanking"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[Esc] cancel action"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_in_txt_action_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "txt file input action"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 replace frame set"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 prepend to frame set"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    3 insert after current frame"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    4 append to frame set"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    5 superimpose frame(s) into frame set"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    6 import palettes only"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[Esc] cancel action"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_in_bmp_action_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "bitmap file input action"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 open as frame color map"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 open as frame set color map"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    3 open as frame mask"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    4 open as frame set mask"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    5 open as display background"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    6 clear display background bitmap"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[Esc] cancel action"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_if_not_z_options()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "signal option if not Z"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 nothing"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 monochrome TTL (on or off)"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    3 monochrome simple average of intensity"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    4 monochrome bitwise OR of intensity"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    5 monochrome weighted rgb gray scale"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    6 monochrome weighted gray for o-scope"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[Esc] cancel action"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_dots_options()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "what to do with dots"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 ignore dots"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 remove dots"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    3 enhance dots by dwell of %d"
                , dwell_on_dot
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[Esc] cancel action"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_out_file_type_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "output file type"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 ild"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 dxf"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    3 wav"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    4 txt"
              );

    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    5 bmp"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    6 wtf"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[ESC] cancel action"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_out_ild_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "output options"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 ild save current frame"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 ild save selected frames"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    3 ild save all frames"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    4 ild generate & save frame effect"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    5 ild generate & save frame set effect"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[ESC] cancel action"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_out_dxf_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "output options"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 dxf save current frame"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 dxf save directory selected frames"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    3 dxf save directory all frames"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[ESC] cancel action"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_out_wav_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "output options"
              );
    y_off += y_line;
    y_off += y_line;
    if(frames_per_second)
        bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "    1 wave animation %3.3f fps of all frames"
                    , frames_per_second
                  );
    else
        bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "    1 wave animation each frame shown once"
                    , frames_per_second
                  );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 wave %6.2f second still of current frame"
                , still_frame_duration
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[ESC] cancel action"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_out_txt_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "output options"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 txt save current frame"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 txt save selected frames"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    3 txt save all frames"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    4 txt save current frame palette"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    5 txt save target palette"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    6 C++ save current frame"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    7 C++ save selected frames"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    8 C++ save all frames"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    9 save system settings (wtf)"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[ESC] cancel action"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_out_bmp_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "output options"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 bmp save %d%c pxl of current frame"
                , output_bmp_size
                , 0xfd
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 bmp save directory %d%c pxl selected frames"
                , output_bmp_size
                , 0xfd
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    3 bmp save directory %d%c pxl all frames"
                , output_bmp_size
                , 0xfd
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[ESC] cancel action"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_value_settings_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "value settings"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 insignificant angle in degrees .......... %3.2lf%c"
                , insignificant_angle * one_radian
                , 0xf8
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 insignificant distance in points ........ %3.2lf"
                , insignificant_distance
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    3 max optimized lit vector in points ...... %3.2lf"
                , lit_delta_max
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    4 max optimized blank vector in points .... %3.2lf"
                , blank_delta_max
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    5 maximum dwell in microseconds ........... %3.2lf"
                , max_dwell_microsec
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    6 lit dwell overhang in vertices .......... %d"
                , lit_dwell_overhang
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    7 what to do with dots .................... %s"
                , LaserBoy_dots_setting_id_to_name(dots_setting).c_str()
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    8 dwell on dot in vertices ................ %d"
                , dwell_on_dot
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    9 move points per key hit ................. %d"
                , displacement_step
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    0 rotate degrees per key hit .............. %3.2f%c"
                , rotation_step * one_radian
                , 0xf8
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    a scale percent per key hit ............... %3.2f%%"
                , magnitude_step * 100.00
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    b rendered arc step in degrees ............ %3.2f%c"
                , rendered_arc_angle
                , 0xf8
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    c frames per second, 0 for once each ...... %3.2lf"
                , frames_per_second
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    d still frame durration in seconds ........ %3.2lf"
                , still_frame_duration
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    e output wave sample rate ................. %d"
                , sample_rate
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    f unframed wave saples per frame .......... %d"
                , unframed_wave_stride
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    g wave signal if not z .................... %s"
                , LaserBoy_wave_signal_id_to_name(signal_if_not_z).c_str()
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    h output bitmap size in square pixels ..... %d"
                , output_bmp_size
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    i display font size factor ................ %d"
                , font_size_factor
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    j background bitmap size factor ........... %d"
                , background_bitmap_size
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.make_rgb(0,0,255), bmp_bg_color, 0, font_size_factor
                , "                                                          %3d"
                , bg_color.b
              );
    bmp_printf(   &bmp, x_off, y_off, bmp.make_rgb(0,255,0), bmp_bg_color, 0, font_size_factor
                , "                                                     %3d"
                , bg_color.g
              );
    bmp_printf(   &bmp, x_off, y_off, bmp.make_rgb(255,0,0), bmp_bg_color, 0, font_size_factor
                , "                                                %3d"
                , bg_color.r
              );
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    k display background color ................"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    l display background gray scale ........... %d"
                , (bg_color.r + bg_color.g + bg_color.b) / 3
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    m onion skin fade ......................... %3.2lf%%"
                , onion_skin_fade * 100.0
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    n onion skin depth ........................ %d"
                , number_of_skins
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    o recolor palette span factor ............. %3.2lf"
                , recolor_span_factor
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    p black limit for color correction  ....... %d"
                , black_level
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[ESC] exit value settings"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_veiw_menu()
{
    int    x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off,
           y_line       =  8 * font_size_factor,
           x_tab        = 30 * y_line,
           x_tab2       =  9 * y_line;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.white
                , bmp_bg_color
                , 0
                , font_size_factor
                , "press & hold view keys, [Shift] reverse"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.make_rgb(255,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "1 ! orbit X   %3.2f%c"
                , rotation_step * one_radian
                , 0xf8
              );
    bmp_printf(   &bmp
                , x_tab
                , y_off
                , bmp.make_rgb(0,255,255)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "5 %% front | back"
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.make_rgb(255,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "2 @ orbit  Y  %3.2f%c"
                , rotation_step * one_radian
                , 0xf8
              );
    bmp_printf(   &bmp
                , x_tab
                , y_off
                , bmp.make_rgb(0,255,255)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "6 ^  left | right"
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.make_rgb(255,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "3 # orbit   Z %3.2f%c"
                , rotation_step * one_radian
                , 0xf8
              );
    bmp_printf(   &bmp
                , x_tab
                , y_off
                , bmp.make_rgb(0,255,255)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "7 &   top | bottom"
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.make_rgb(255,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "4 $ orbit XYZ %3.2f%c"
                , rotation_step * one_radian
                , 0xf8
              );
    bmp_printf(   &bmp
                , x_tab
                , y_off
                , bmp.make_rgb(0,255,255)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "8 * orbit & frame"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.make_rgb(255,0,255)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "q Q pan X  %5d points"
                , displacement_step
              );
    bmp_printf(   &bmp
                , x_tab
                , y_off
                , bmp.make_rgb(0,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "a A zoom X  %3.2f%%"
                , magnitude_step * 100.00
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.make_rgb(255,0,255)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "w W pan  Y %5d points"
                , displacement_step
              );
    bmp_printf(   &bmp
                , x_tab
                , y_off
                , bmp.make_rgb(0,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "s S zoom  Y %3.2f%%"
                , magnitude_step * 100.00
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.make_rgb(255,0,255)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "e E pan XY %5d points"
                , displacement_step
              );
    bmp_printf(   &bmp
                , x_tab
                , y_off
                , bmp.make_rgb(0,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "d D zoom XY %3.2f%%"
                , magnitude_step * 100.00
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.make_rgb(255,0,255)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "r R center"
              );
    bmp_printf(   &bmp
                , x_tab
                , y_off
                , bmp.make_rgb(0,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "f F 1:1"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.white
                , bmp_bg_color
                , 0
                , font_size_factor
                , "current view"
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_tab2
                , y_off
                , bmp.make_rgb(255,0,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "X"
              );
    bmp_printf(   &bmp
                , x_tab2 * 2
                , y_off
                , bmp.make_rgb(0,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "Y"
              );
    bmp_printf(   &bmp
                , x_tab2 * 3
                , y_off
                , bmp.make_rgb(0,0,255)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "Z"
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.make_rgb(255,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , " angle                               degrees"
              );
    bmp_printf(   &bmp
                , x_tab2
                , y_off
                , bmp.make_rgb(255,255,0)
                , bmp.make_rgb(80,0,0)
                , 0
                , font_size_factor
                , "%-3.4f"
                , view_angle.x * one_radian
              );
    bmp_printf(   &bmp
                , x_tab2 * 2
                , y_off
                , bmp.make_rgb(255,255,0)
                , bmp.make_rgb(0,80,0)
                , 0
                , font_size_factor
                , "%-3.4f"
                , view_angle.y * one_radian
              );
    bmp_printf(   &bmp
                , x_tab2 * 3
                , y_off
                , bmp.make_rgb(255,255,0)
                , bmp.make_rgb(0,0,80)
                , 0
                , font_size_factor
                , "%-3.4f"
                , view_angle.z * one_radian
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.make_rgb(255,0,255)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "offset                               points"
              );
    bmp_printf(   &bmp
                , x_tab2
                , y_off
                , bmp.make_rgb(255,0,255)
                , bmp.make_rgb(80,0,0)
                , 0
                , font_size_factor
                , "%-8.0f"
                , view_offset.x
              );
    bmp_printf(   &bmp
                , x_tab2 * 2
                , y_off
                , bmp.make_rgb(255,0,255)
                , bmp.make_rgb(0,80,0)
                , 0
                , font_size_factor
                , "%-8.0f"
                , view_offset.y
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.make_rgb(0,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , " scale"
              );
    bmp_printf(   &bmp
                , x_tab2
                , y_off
                , bmp.make_rgb(0,255,0)
                , bmp.make_rgb(80,0,0)
                , 0
                , font_size_factor
                , "%-3.4f"
                , view_scale.x
              );
    bmp_printf(   &bmp
                , x_tab2 * 2
                , y_off
                , bmp.make_rgb(0,255,0)
                , bmp.make_rgb(0,80,0)
                , 0
                , font_size_factor
                , "%-3.4f"
                , view_scale.y
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.make_rgb(255,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "      o O apply angle  to frame | all"
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.make_rgb(255,0,255)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "      t T apply offset to frame | all"
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.make_rgb(0,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "      g G apply scale  to frame | all"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.white
                , bmp_bg_color
                , 0
                , font_size_factor
                , "  [Esc] exit view & apply"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_visuals_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "user interface visuals settings"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((show_vertices)?(" --> 1 vertices"):("     1 vertices"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((show_blanking)?(" --> 2 blanking"):("     2 blanking"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((show_palette)?(" --> 3 palette"):("     3 palette"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((show_target_palette)?(" --> 4 target palette"):("     4 target palette"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((show_stats)?(" --> 5 stats"):("     5 stats"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((show_fixed_bounds)?(" --> 6 fixed bounds"):("     6 fixed bounds"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((show_fixed_origin)?(" --> 7 fixed origin"):("     7 fixed origin"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((show_floating_axis)?(" --> 8 floating axis"):("     8 floating axis"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((show_floating_bounds)?(" --> 9 floating bounds"):("     9 floating bounds"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((show_onion_skin)?(" --> 0 onion skin"):("     0 onion skin"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((show_onion_visuals)?(" --> a onion visuals"):("     a onion visuals"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((show_TUI_clue)?(" --> b TUI clue"):("     b TUI clue"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "     c show | hide all visuals"
              );
    //------------------------------------------------------------------------
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[Esc] exit user interface visuals settings"
              );
    //------------------------------------------------------------------------
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_switch_settings_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);
    //------------------------------------------------------------------------
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "switch settings"
              );
    //------------------------------------------------------------------------
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "animation"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((loop_animation)?(" --> 1 loop animation"):("     1 loop animation"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((approximate_frame_rate)?(" --> 2 animate at approximated frame rate"):("     2 animate at approximated frame rate"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((show_effects_generation)?(" --> 3 show effect rendering"):("     3 show effect rendering"))
              );
    //------------------------------------------------------------------------
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "input"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((auto_scale_real)?(" --> 4 auto scale real coordinate import"):("     4 auto scale real coordinate import"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((maintain_real_origin)?(" --> 5 maintain real coordinate origin"):("     5 maintain real coordinate origin"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((z_from_unformatted)?(" --> 6 import Z from unformatted wave"):("     6 import Z from unformatted wave"))
              );
    //------------------------------------------------------------------------
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "optimization"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((no_equivalent_vectors)?(" --> 7 remove equal vectors"):("     7 remove equal vectors"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((black_dwell_vertices)?(" --> 8 black dwell"):("     8 black dwell"))
              );
    //------------------------------------------------------------------------
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "color"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((no_black_match)?(" --> 9 no black palette match"):("     9 no black palette match"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((allow_lit_black)?(" --> 0 allow lit black"):("     0 allow lit black"))
              );
    //------------------------------------------------------------------------
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "output"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((auto_optimize)?(" --> a optimize before save"):("     a optimize before save"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                   , ((dumb_file_format)?(" --> b save ild format 4 or 5"):("     b save ild format 4 or 5"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                   , ((save_true_color_dxf)?(" --> c save dxf in true color"):("     c save dxf in true color"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((auto_flatten_z)?(" --> d flatten Z for wave output"):("     d flatten Z for wave output"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((invert_wave_output)?(" --> e invert wave output"):("     e invert wave output"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((auto_apply_offsets)?(" --> f apply system offsets to wave"):("     f apply system offsets to wave"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                   , ((channels_of_wav_out == 8)?(" --> g save waves with added silence (8ch)"):("     g save waves with added silence (8ch)"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                   , ((save_txt_with_color)?(" --> h save txt frames with integrated color"):("     h save txt frames with integrated color"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                   , ((save_txt_unit)?(" --> i save txt in unit coordinates (-1.0 to +1.0)"):("     i save txt in unit coordinates (-1.0 to +1.0)"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                   , ((save_txt_color_hex)?(" --> j save txt colors as hex"):("     j save txt colors as hex"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                   , ((save_txt_named_palettes)?(" --> k save txt named LaserBoy palette values"):("     k save txt named LaserBoy palette values"))
              );
    //------------------------------------------------------------------------
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[Esc] exit switch settings"
              );
    //------------------------------------------------------------------------
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_palette_tranz_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "palette transforms (to frames)"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  1 2 3 span X | Y | Z"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    4 5 span radially | angular from origin"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    6 7 span | index vectors"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    8 9 span | index segments"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    0 a random color vectors | segments"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    b c set to target by best match | index"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  d e f strip color white | logical OR | gray scale"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  g h i 24-bit to palette lossless | bits 332 | best"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      j promote to 24-bit"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[Shift] key from above applies to all frames"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      k best reduce all frames to 1 palette"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "palette transforms (to palettes and the set)"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      l omit unused added palettes"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    m M copy current | target"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      n add new target"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    o O add new color to current | target"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    p P +-selected color index"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    q Q set selected color in current | target"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    r R insert color before selected in current | target"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    s S linear | circular blend upto 6 colors in target"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    t T omit color selected in current | target"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    u U omit selected color to end of current | target"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      v omit target"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "  [Esc] exit palette transforms"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_frame_set_tranz_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "frame set transforms"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      1 omit selected"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      2 trim to selected"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      3 omit repeats from wave import"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      4 copy selected to begining"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      5 copy selected to end"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      6 move selected to begining"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      7 move selected to end"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      8 add empty frame to begining"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      9 add empty frame after current"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      0 add empty frame to end"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      a reverse frame set order"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      b reverse order of selected frames"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      c make current frame first"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      d randomize frame set order"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      e omit every n-th frame"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      > explode frame into segment frame set"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      < collapse selected frames into one"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      ! explode all frames into segment frame sets"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "[Space] select | unselect frame"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[Enter] show | hide help"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "  [Esc] exit frame set transforms"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_frame_tranz_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "frame transforms"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 2 flatten Z | Z order"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    3 4 mirror | flip"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    5 6 quarter turn CW | CCW"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      7 reverse vector order"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      8 impose bit resolution"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      9 black to blank"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      0 blank to black"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      a impose black level"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      b convert to dots"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      c remove blank span"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      d remove angle dwell"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      e remove dots (not lines)"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      f reduce lit vectors"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      g blank equivalent vectors"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      h bond segments (reorder)"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      i minimize"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      j fracture segments"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      k randomize segments"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      l reorder segments"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      m add angle dwell"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      n add lit distance span"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      o add blank distance span"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      p enhance dots"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      q optimize"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[Shift] key applies to all frames"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[Enter] show | hide help"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "  [Esc] exit frame transforms"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_draw_color_blank_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "draw, color, blank"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "   [{ ]} +-vertex  %c | %c"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "   iI oO +-segment %c | %c"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    u  U +-segment %c-%c"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    \\  |           %c=0 | %c=end"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "xX yY zZ move      %c X | Y | Z  +-%d"
                , 0x0f
                , displacement_step
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , " q  w  e set       %c X | Y | Z"
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "   fF gG rotate    %c Z | Y axis +-%3.2f%c"
                , 0x0f
                , rotation_step * one_radian
                , 0xf8
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    m  M magnitude %c            +-%3.2f%%"
                , 0x0f
                , magnitude_step * 100.00
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    j  J assign    %c=%c | %c=%c"
                , 0x0f
                , 0x01
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "     a A unblank all vectors | frames"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "     b B blank | unblank %c-%c"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "     n N bond black | unbond unblack %c-%c"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "     p P +-selected color"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "     c C set %c-%c to selected color | get %c color"
                , 0x01
                , 0x0f
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "     - _ set %c-%c to palette by index | span"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "     h H +-cycle colors in %c-%c"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "       ~ set black in %c-%c to selected color"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "     = + set black in %c-%c to palette by index | span"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "     l L draw line | render text %c-%c"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "     d D radius %c-%c crcl | elpt arc %3.2f%c steps"
                , 0x01
                , 0x0f
                , rendered_arc_angle
                , 0xf8
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "       k radius %c-%c polygon"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "     , < make %c | %c first vertex"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "       ' convert %c-%c to dots"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "     . > add vector after %c | add blank at %c"
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , " [Bkspc] omit %c"
                , 0x0f
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "       ` copy frame"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , " [Enter] show | hide help"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "   [Esc] exit draw, color, blank"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_move_scale_rotate_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "move, scale, rotate mode entered"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "   [{ ]}         %c | %c +-vertex"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "   iI oO         %c | %c +-segment"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    u  U         %c-%c   +-segment"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    \\  |         %c=0 | %c=end"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "xX yY zZ move    %c-%c X | Y | Z             +-%d"
                , 0x01
                , 0x0f
                , displacement_step
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , " q  w  e set     %c-%c X | Y | Z     center"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "   fF gG rotate  %c-%c Z | Y axis on center  +-%3.2f%c"
                , 0x01
                , 0x0f
                , rotation_step * one_radian
                , 0xf8
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "9( 0) -_ scale   %c-%c X | Y | Z  on center  +-%3.2f%%"
                , 0x01
                , 0x0f
                , magnitude_step * 100.00
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    =  + scale   %c-%c    XYZ     on center  +-%3.2f%%"
                , 0x01
                , 0x0f
                , magnitude_step * 100.00
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "aA sS dD move fulcrum X | Y | Z +-%d"
                , displacement_step
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , " j  k  l set  fulcrum X | Y | Z"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    ,  . move fulcrum  to %c | %c"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "   rR tT rotate  %c-%c Z | Y axis on fulcrum +-%3.2f%c"
                , 0x01
                , 0x0f
                , rotation_step * one_radian
                , 0xf8
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "vV bB nN scale   %c-%c X | Y | Z  on fulcrum +-%3.2f%%"
                , 0x01
                , 0x0f
                , magnitude_step * 100.00
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    m  M scale   %c-%c    XYZ     on fulcrum +-%3.2f%%"
                , 0x01
                , 0x0f
                , magnitude_step * 100.00
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "       < reverse %c-%c vector order"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "       c copy    %c-%c (in place) to end"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "       ` copy frame"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , " [Enter] show | hide help"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "   [ESC] exit move, scale, rotate mode"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_wave_utilities_menu()
{
    int y_line = 8 * font_size_factor,
        x_off  = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "wave file utilities"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 wave show screen settings"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 show wave on screen"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    3 show LaserBoy wave header information"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    4 show unformatted wave header information"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    5 set system wave sample offsets"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    6 apply offsets to wave"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    7 overwrite offsets in LaserBoy wave header"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    8 invert global polarity of wave"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    9 invert signal polarities of LaserBoy wave"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    0 clear signal polarity list in LaserBoy wave"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    a impose 8-bit color and black level of %d"
                , black_level
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    b impose bit resolution on wave"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    c split wave into XY, r, g, b, i, LR"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    d join XY, r, g, b, i waves to 6 channel wave"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    e join XY, rg, b, waves to 6 channel wave"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    f join XY, rg, b, LR waves to 8 channel wave"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    g add stereo audio to wave (6 to 8)"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[ESC] exit wave file utilities"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_show_wave_options_menu()
{
    int y_line = 8 * font_size_factor,
        x_off  = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "show wave options"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 show LaserBoy wave"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 show LaserBoy wave as unformatted"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    3 show unformatted wave"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[ESC] exit show wave options"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_wave_show_settings_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "wave show settings"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((wave_show_vertices)?(" --> 1 vertices"):("     1 vertices"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((wave_show_blanks)?(" --> 2 blanking"):("     2 blanking"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((wave_real_time)?(" --> 3 real time"):("     3 real time"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((wave_show_inverted)?(" --> 4 inverted"):("     4 inverted"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((wave_clean_screen)?(" --> 5 clean screen"):("     5 clean screen"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((wave_loop)?(" --> 6 loop"):("     6 loop"))
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[Esc] exit wave show settings"
              );
    return;
}

//############################################################################
bool LaserBoy_TUI::overlay_wave_stats(string file_name)
{
    int       i,
              y_line       = 8 * font_size_factor,
              x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
              y_off        = x_off;
    u_int     bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);
    fstream   in;
    LaserBoy_wave_header header;
    //------------------------------------------------------------------------
    in.open(file_name.c_str(), ios::in | ios::binary);
    if(in.is_open())
    {
        header.from_fstream_wave(in);
        in.close();
    }
    else
        return false;
    //------------------------------------------------------------------------
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "wave stats                                              "
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , file_name.c_str()
              );
    y_off += y_line;
    y_off += y_line;
    if(header.version != "!LaserBoy!")
        bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "LaserBoy wave header version : %s"
                    , header.version.c_str()
                  );
    else
        bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "wave has no LaserBoy format information"
                  );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "num_samples ..........: %d"
                , header.num_samples
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "sample_rate ..........: %d"
                , header.sample_rate
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "number of channels ...: %d"
                , header.num_channels
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "bits per sample ......: %d"
                , header.bits_per_sample
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "playing time .........: %s hh:mm:ss.ss"
                , time_as_string((double)header.num_samples / header.sample_rate).c_str()
              );
    if(header.version != "!LaserBoy!")
    {
        y_off += y_line;
        y_off += y_line;
        bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , ((header.LaserBoy_wave_mode & LASERBOY_WAVE_POSITIVE) ? ("wave polarity ........: positive signal") : ("wave polarity ........: negative signal"))
                  );
        if(header.LaserBoy_wave_mode & LASERBOY_WAVE_OPTIMIZED)
        {
            y_off += y_line;
            bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                        , "lit dwell overhang ...: %d"
                        , header.parms.lit_dwell_overhang
                      );
            y_off += y_line;
            bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                        , "lit distance delta max: %6.2f"
                        , header.parms.lit_delta_max
                      );
            y_off += y_line;
            bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                        , "blank dist delta max .: %6.2f"
                        , header.parms.blank_delta_max
                      );
            y_off += y_line;
            bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                        , "longest dwell u_sec ..: %6.2f"
                        , header.parms.max_dwell_microsec
                      );
            y_off += y_line;
            bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                        , "insignificant distance: %6.2f"
                        , header.parms.insignificant_distance
                      );
            y_off += y_line;
            bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                        , "insignificant angle ..: %6.2f"
                        , header.parms.insignificant_angle
                      );
            y_off += y_line;
            bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                        , "frames per second ....: %6.2f"
                        , header.parms.frames_per_second
                      );
        }
        y_off += y_line;
        y_off += y_line;
        bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                    , "ch: sig : contents     : LSB subcode  : offset : res"
                  );
        for(i = 0; i < header.num_channels; i++)
        {
            y_off += y_line;
            bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                        , "%2d: %s : %s : %s :   %4d : %3d"
                        , i
                        , ((header.signal_id[i] >= 0) ? ("   ") : ("neg"))
                        , LaserBoy_wave_signal_id_to_name(header.signal_id[i]).c_str()
                        , LaserBoy_wave_LSB_tag_to_name(header.LSB_tag[i]).c_str()
                        , header.offset[i]
                        , header.resolution[i]
                      );
        }
    }
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "press any key to continue"
              );
    return true;
}

//############################################################################
short LaserBoy_TUI::overlay_wave_signal_polarities(string file_name, int polarity[8])
{
    int            i,
                   y_line       = 8 * font_size_factor,
                   x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
                   y_off        = x_off;
    u_int          bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);
    fstream        in;
    //------------------------------------------------------------------------
    in.open(file_name.c_str(), ios::in | ios::binary);
    LaserBoy_wave_header header(in);
    in.close();
    //------------------------------------------------------------------------
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "wave stats                                          "
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , file_name.c_str()
              );
    if(header.version != "!LaserBoy!")
    {
        y_off += y_line;
        y_off += y_line;
        bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , ((header.LaserBoy_wave_mode & LASERBOY_WAVE_POSITIVE) ? ("wave polarity ........: positive signal") : ("wave polarity ........: negative signal"))
                  );
    }
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "ch: sig : contents"
              );
    for(i = 0; i < header.num_channels; i++)
    {
        y_off += y_line;
        bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "%2d: %s : %s"
                    , i
                    , ((header.signal_id[i] >= 0) ? ("   ") : ("neg"))
                    , LaserBoy_wave_signal_id_to_name(header.signal_id[i]).c_str()
                  );
    }
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "wave signal polarity matrix"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [0] X          %s"
                , ((polarity[0] == -1) ? ("") : ((polarity[0]) ? ("invert") : ("unchanged")))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [1] Y          %s"
                , ((polarity[1] == -1) ? ("") : ((polarity[1]) ? ("invert") : ("unchanged")))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [2] red        %s"
                , ((polarity[2] == -1) ? ("") : ((polarity[2]) ? ("invert") : ("unchanged")))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [3] green      %s"
                , ((polarity[3] == -1) ? ("") : ((polarity[3]) ? ("invert") : ("unchanged")))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [4] blue       %s"
                , ((polarity[4] == -1) ? ("") : ((polarity[4]) ? ("invert") : ("unchanged")))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [5] Z or other %s"
                , ((polarity[5] == -1) ? ("") : ((polarity[5]) ? ("invert") : ("unchanged")))
              );
    if(header.num_channels == 8)
    {
        y_off += y_line;
        bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "  [6] L audio    %s"
                    , ((polarity[6] == -1) ? ("") : ((polarity[6]) ? ("invert") : ("unchanged")))
                  );
        y_off += y_line;
        bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "  [7] R audio    %s"
                    , ((polarity[7] == -1) ? ("") : ((polarity[7]) ? ("invert") : ("unchanged")))
                  );
    }
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[Esc] cancel"
              );
    return header.num_channels;
}

//############################################################################
void LaserBoy_TUI::overlay_wave_offsets()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "wave sample offsets"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [0] X          %d"
                , wave_offsets[0]
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [1] Y          %d"
                , wave_offsets[1]
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [2] red        %d"
                , wave_offsets[2]
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [3] green      %d"
                , wave_offsets[3]
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [4] blue       %d"
                , wave_offsets[4]
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [5] Z or other %d"
                , wave_offsets[5]
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [6] L audio    %d"
                , wave_offsets[6]
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [7] R audio    %d"
                , wave_offsets[7]
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[Esc] cancel"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_bit_resolution_settings()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "signal bit resolution settings"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [0] X       %d"
                , 16 - signal_bit_mask[0]
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [1] Y       %d"
                , 16 - signal_bit_mask[1]
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [5] Z       %d"
                , 16 - signal_bit_mask[5]
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [2] red     %d"
                , 16 - signal_bit_mask[2]
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [3] green   %d"
                , 16 - signal_bit_mask[3]
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [4] blue    %d"
                , 16 - signal_bit_mask[4]
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [6] L audio %d"
                , 16 - signal_bit_mask[6]
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [7] R audio %d"
                , 16 - signal_bit_mask[7]
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[Esc] cancel"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_file_list(string directory)
{
    bool            found_one    = false;
    int             i,
                    y_line       = 8 * font_size_factor,
                    x_off        = bmp.yres + LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
                    y_off        = y_line;
    u_int           bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);
    DIR            *pDIR;
    struct dirent  *pDirEnt;

    pDIR = opendir(directory.c_str());
    if(pDIR == NULL)
        return;
    bmp_printf(&bmp,
               x_off,
               0,
               bmp.white,
               bmp_bg_color,
               0,
               font_size_factor,
               "%s                  ",
               directory.c_str()
              );
    pDirEnt = readdir(pDIR);
    while(pDirEnt != NULL)
    {
        if(    strcmp(pDirEnt->d_name, ".")  // not the dot
            && strcmp(pDirEnt->d_name, "..") // not the dot dot
          )
        {
            bmp_printf(&bmp,
                       x_off,
                       y_off,
                       next_hue(),
                       bmp_bg_color, 0,
                       font_size_factor,
                       "%s                  ",
                       pDirEnt->d_name
                      );
            y_off += y_line;
            found_one = true;
        }
        pDirEnt = readdir(pDIR);
    }
    closedir(pDIR);
    if(!found_one)
    {
        bmp_printf(&bmp,
                   x_off,
                   y_off,
                   bmp.white,
                   bmp_bg_color,
                   0,
                   font_size_factor,
                   "none found        "
                  );
        y_off += y_line;
    }
    bmp_printf(&bmp,
               x_off,
               y_off,
               bmp.white,
               bmp_bg_color,
               0,
               font_size_factor,
               "                  "
              );
    y_off += y_line;
    bmp_printf(&bmp,
               x_off,
               y_off,
               bmp.white,
               bmp_bg_color,
               0,
               font_size_factor,
               "                  "
              );
    y_off += y_line;
    bmp_printf(&bmp,
               x_off,
               y_off,
               bmp.white,
               bmp_bg_color,
               0,
               font_size_factor,
               "                  "
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_directory_list(string directory)
{
    bool            found_one    = false;
    int             i,
                    y_line       = 8 * font_size_factor,
                    x_off        = bmp.yres + LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
                    y_off        = y_line;
    u_int           bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);
    DIR            *pDIR;
    struct dirent  *pDirEnt;

    pDIR = opendir(directory.c_str());
    if(pDIR == NULL)
        return;
    bmp_printf(&bmp,
               x_off,
               0,
               bmp.white,
               bmp_bg_color,
               0,
               font_size_factor,
               "%s                  ",
               directory.c_str()
              );
    pDirEnt = readdir(pDIR);
    while(pDirEnt != NULL)
    {
        if(!strstr(pDirEnt->d_name, ".")) // if there is no dot in the name
        {
            bmp_printf(&bmp,
                       x_off,
                       y_off,
                       next_hue(),
                       bmp_bg_color,
                       0,
                       font_size_factor,
                       "%s                  ",
                       pDirEnt->d_name
                      );
            y_off += y_line;
            found_one = true;
        }
        pDirEnt = readdir(pDIR);
    }
    closedir(pDIR);
    if(!found_one)
    {
        bmp_printf(&bmp,
                   x_off,
                   y_off,
                   bmp.white,
                   bmp_bg_color,
                   0,
                   font_size_factor,
                   "none found        "
                  );
        y_off += y_line;
    }
    bmp_printf(&bmp,
               x_off,
               y_off,
               bmp.white,
               bmp_bg_color,
               0,
               font_size_factor,
               "                  "
              );
    y_off += y_line;
    bmp_printf(&bmp,
               x_off,
               y_off,
               bmp.white,
               bmp_bg_color,
               0,
               font_size_factor,
               "                  "
              );
    y_off += y_line;
    bmp_printf(&bmp,
               x_off,
               y_off,
               bmp.white,
               bmp_bg_color,
               0,
               font_size_factor,
               "                  "
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_frame_effects()
{
    int    i,
           y_line       = 8 * font_size_factor,
           x_off        = bmp.yres + LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(&bmp,
               x_off,
               0,
               bmp.white,
               bmp_bg_color,
               0,
               font_size_factor,
               "frame effects             "
              );
    if(frame_effect_names.size() > 0)
    {
        for(i = 0; i < (int)frame_effect_names.size(); i++)
        {
            y_off += y_line;
            bmp_printf(&bmp,
                       x_off,
                       y_off,
                       next_hue(),
                       bmp_bg_color,
                       0,
                       font_size_factor,
                       "%s                 ",
                       frame_effect_names[i].c_str()
                      );
        }
    }
    else
    {
        y_off += y_line;
        bmp_printf(&bmp,
                   x_off,
                   y_off,
                   next_hue(),
                   bmp_bg_color,
                   0,
                   font_size_factor,
                   "no frame effects loaded"
                  );
    }
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_frame_set_effects()
{
    int    i,
           y_line       = 8 * font_size_factor,
           x_off        = bmp.yres + LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(&bmp,
               x_off,
               0,
               bmp.white,
               bmp_bg_color,
               0,
               font_size_factor,
               "frame set effects          "
              );
    if(frame_set_effect_names.size() > 0)
    {
        for(i = 0; i < (int)frame_set_effect_names.size(); i++)
        {
            y_off += y_line;
            bmp_printf(&bmp,
                       x_off,
                       y_off,
                       next_hue(),
                       bmp_bg_color,
                       0,
                       font_size_factor,
                       "%s                 ",
                       frame_set_effect_names[i].c_str()
                      );
        }
    }
    else
    {
        y_off += y_line;
        bmp_printf(&bmp,
                   x_off,
                   y_off,
                   next_hue(),
                   bmp_bg_color,
                   0,
                   font_size_factor,
                   "no frame set effects loaded"
                  );
    }
    return;
}

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
